C**********************************************************************C
C***********************  BeoPEST Public Data  ************************C
C**********************************************************************C
      MODULE BEOPEST
      IMPLICIT NONE
      LOGICAL BEOMASTER,BEOSLAVE,BEOLASTLOC
      INTEGER MAXNODE
      DOUBLE PRECISION last_first_obs                                !jd
      INTEGER PARLAM_FLAG                                            !jd
      DOUBLE PRECISION AVERAGE_WALLTIME                              !jd
      END
C**********************************************************************C
C***********************  BeoPEST private data  ***********************C
C**********************************************************************C
C  BeoPEST/TCP uses LIVE to track nodes
C  LIVE
C    -1 = idle
C     0 = offline
C     1 = running
C  BeoPEST/MPI uses IRDQ to store requests
      MODULE BEOPRIV
      IMPLICIT NONE
      ALLOCATABLE      LIVE,IRDQ,WALL,STRT,NUM,TEXT,PARBUF,OBSBUF
      allocatable      oldtext,super_needed                          !jd
      allocatable      suspect                                       !jd
      LOGICAL          TCP,MPI
      INTEGER          NPAR,NOBS
      INTEGER          LIVE(:),IRDQ(:,:),NUM(:)
      INTEGER          super_needed(:)                               !jd
      integer          suspect(:)                                    !jd
      DOUBLE PRECISION WALL(:),STRT(:),TOTIM,TSUM
      DOUBLE PRECISION PARBUF(:,:),OBSBUF(:,:)
      CHARACTER*256    TEXT(:)
      character*80     oldtext(:)                                    !jd
      character*500    charbuf                                       !jd
      integer          ient_avbuf                                    !jd
      integer          nument_avbuf                                  !jd
      integer          av_operating                                  !jd
      double precision, allocatable :: avbuf(:)                      !jd

      END

C -- I added the following module myself because I get nervous about      !jd
C    allocating and de-allocating. I don't trust compilers.               !jd

      MODULE BEORUN_DATA                                                  !jd
        IMPLICIT NONE                                                     !jd
        INTEGER                          IALLOC_BEORUN_DATA               !jd
        DOUBLE PRECISION, ALLOCATABLE :: PAR(:)                           !jd
        INTEGER, ALLOCATABLE          :: MODE(:),STAT(:)                  !jd
        INTEGER                          PARDIM,MODEDIM,STATDIM           !jd
      END                                                                 !jd


C**********************************************************************C
C********************  Initialize communications  *********************C
C**********************************************************************C
      SUBROUTINE BEOINIT(HOST,MXNODE)
      USE           BEOPEST
      INCLUDE       'beopest.inc'
#ifdef BEOMPI
      INCLUDE 'mpif.h'
      INTEGER IOS
#endif
      INTEGER          MXNODE,NODE
      DOUBLE PRECISION WALLTIME
      CHARACTER*10     CNN
      CHARACTER*(*)    HOST

      TCP = (MXNODE.gt.0)
      MPI = .not.TCP
C  Initialize TCP
      if (TCP) then
        MAXNODE = MXNODE
        call TCPOPEN(MAXNODE,TRIM(HOST)//CHAR(0),NODE)
C  Initialize MPI
      else
#ifdef BEOMPI
        call MPI_INIT(IOS)
        call MPI_COMM_RANK(MPI_COMM_WORLD,NODE   ,IOS)
        call MPI_COMM_SIZE(MPI_COMM_WORLD,MAXNODE,IOS)
        if (MAXNODE.eq.1) STOP 'Cannot run BeoPEST/MPI with one node'
#else
        STOP 'Compiled without MPI'
#endif
      endif

C  Mark master and slave
      BEOMASTER = (NODE.eq.0)
      BEOSLAVE  = (NODE.gt.0)
C  Allocate memory for tracking slave execution times
      if (BEOMASTER) then
        ALLOCATE(WALL(MAXNODE),STRT(MAXNODE),NUM(MAXNODE),
     +  TEXT(MAXNODE),oldtext(maxnode),super_needed(maxnode),
     +  suspect(maxnode))
        oldtext=' '      ! an array                                       !jd
        NUM  = 0
        WALL = 0
        super_needed=0
        suspect=0                                                         !jd
        nument_avbuf=0                                                    !jd
        ient_avbuf=0                                                      !jd
        av_operating=0                                                    !jd
      else
        MAXNODE = 1
      endif

C  Initialize MASTER array for tracking slaves
      if (BEOMASTER) then
        TSUM = 0
        TOTIM = WALLTIME()
        if (TCP) then
          ALLOCATE(LIVE(MAXNODE))
          LIVE = 0
        else
#ifdef BEOMPI
          ALLOCATE(IRDQ(2,MAXNODE))
          IRDQ = MPI_REQUEST_NULL
C         Directory name for each slave
          do NODE=1,MAXNODE
            WRITE(CNN,'(I10)') NODE
            TEXT(NODE) = TRIM(HOST)//TRIM(ADJUSTL(CNN))
          enddo
#endif
        endif
C  On MPI slaves change directory
      else if (MPI) then
        WRITE(CNN,'(I10)') NODE
        call SETDIR(TRIM(HOST)//TRIM(ADJUSTL(CNN))//CHAR(0))
      endif
      END
C**********************************************************************C
C***********************  Get number of nodes  ************************C
C**********************************************************************C
      SUBROUTINE BEONODE(KP0,KP1)
      USE     BEOPEST
      use pestdata, only : irmr                                         !jd
      INCLUDE 'beopest.inc'
      INTEGER KP0,KP1
      logical lopened                                                   !jd
      integer  dd,mm,yy,hh,min,j,timeflag,jj                            !jd
      real     ss                                                       !jd
      character*10 anew                                                 !jd
      character*11 bresult1,bresult2,bresult                            !jd

      timeflag=0                                                        !jd
      KP0 = 1
      if (TCP) then
        call TCPNODE(LIVE,WALL,TEXT,KP1)
      else
        KP1 = MAXNODE-1
      endif
      if(kp1.ge.1) then
        inquire(unit=irmr,opened=lopened)                               !jd
        if(lopened)then                                                 !jd
          do j=1,kp1                                                    !jd
            jj=len(oldtext(j))                                          !jd
            if(text(j)(1:jj).ne.oldtext(j)(1:jj))then                   !jd
              if(timeflag.eq.0)then                                     !jd
                call date_and_time(date=bresult1,time=bresult2)         !jd
C               read(bresult1,'(i4,i2,i2)') yy,mm,dd                    !jd
                read(bresult2,'(i2,i2,f6.0)') hh,min,ss                 !jd
                write(bresult,'(i2.2,'':'',i2.2,'':'',f5.2)') hh,min,ss !jd
                if(bresult(7:7).eq.' ')bresult(7:7)='0'                 !jd
                timeflag=1                                              !jd
              end if                                                    !jd
              call writint(anew,j)                                      !jd
              write(irmr,10) bresult,trim(anew),trim(text(j))           !jd
10            format(3x,a,':- index of ',a,                             !jd
     +        ' assigned to node at ',                                  !jd
     +        'working directory "',a,'".')                             !jd
              oldtext(j)=text(j)                                        !jd
              write(6,20) trim(text(j))                                 !jd
20            format(/,'    New slave has appeared:-',/,                !jd
     +        '    Local working directory "',a,'"',/)                  !jd
            end if                                                      !jd
          end do                                                        !jd
#ifdef FLUSHFILE
          call flush(irmr)                                              !jd
#endif
        end if                                                          !jd
      end if                                                            !jd
      END
C**********************************************************************C
C**********************  Terminate connections  ***********************C
C**********************************************************************C
      SUBROUTINE BEOFIN
      USE      BEOPEST
      INCLUDE 'beopest.inc'
#ifdef BEOMPI
      INCLUDE 'mpif.h'
      INTEGER IOS
#endif
      INTEGER K,KP0,KP1
      integer ierr                                                    !jd
      DOUBLE PRECISION WALLTIME
      CHARACTER*256 BEOTEXT

C  Master sends count zero (close) to slaves before closing TCP connections
      if (BEOMASTER) then
        call BEONODE(KP0,KP1)
        do K=KP0,KP1
          print '(A,I4,F8.3,I5,1X,A)','Node',
     |      K,WALL(K),NUM(K),TRIM(BEOTEXT(K))
        enddo
        TOTIM = WALLTIME()-TOTIM
        print '(A,F12.1)','Total CPU time     ',TSUM
        print '(A,F12.1)','Total elapsed time ',TOTIM
        print '(A,F12.3)','Speedup            ',TSUM/TOTIM
        if (TCP) then
          do K=KP0,KP1
            if (LIVE(K).ne.0) call TCPSEND(K,0,4,1)
          enddo
          call TCPCLOSE
          DEALLOCATE(LIVE)
        else
#ifdef BEOMPI
          do K=KP0,KP1
            call MPI_SEND(0,1,MPI_INTEGER,K,0,MPI_COMM_WORLD,IOS)
          enddo
          call MPI_Finalize(IOS)
          DEALLOCATE(IRDQ)
#endif
        endif
        DEALLOCATE(WALL,STRT,NUM,TEXT,PARBUF,OBSBUF,stat=ierr)        !jd
        deallocate(oldtext,super_needed,stat=ierr)                    !jd
        deallocate(suspect,stat=ierr)                                 !jd
        deallocate(avbuf,stat=ierr)                                   !jd
C  Slave just closes connection
      else if (BEOSLAVE) then
        if (TCP) then
          call TCPCLOSE
        else
#ifdef BEOMPI
          call MPI_Finalize(IOS)
#endif
        endif
      endif
      END
C**********************************************************************C
C********************  Broadcast Parameter Counts  ********************C
C**********************************************************************C
      SUBROUTINE BEOCAST(NP,NT,CSUM,SV)
      USE     BEOPEST
      INCLUDE 'beopest.inc'
#ifdef BEOMPI
      INCLUDE 'mpif.h'
#endif
      INTEGER NP,NT,CSUM,SV,N,IOS
      CHARACTER*8 AVERSION,BVERSION

      INCLUDE 'version.inc'
      BVERSION = AVERSION

C  MPI Broadcast values
      if (MPI) then
#ifdef BEOMPI
        call MPI_BCAST(BVERSION,8,MPI_CHARACTER,0,MPI_COMM_WORLD,IOS)
        call MPI_BCAST(NP,1,MPI_INTEGER,0,MPI_COMM_WORLD,IOS)
        call MPI_BCAST(NT,1,MPI_INTEGER,0,MPI_COMM_WORLD,IOS)
        call MPI_BCAST(CSUM,1,MPI_INTEGER,0,MPI_COMM_WORLD,IOS)
        call MPI_BCAST(SV,1,MPI_INTEGER,0,MPI_COMM_WORLD,IOS)
#endif
C  Set values to send on master
      else if (BEOMASTER) then
        call TCPCAST(BVERSION,NP,NT,CSUM,SV)
C  Receive values on slave
      else if (BEOSLAVE) then
        call TCPRECV(0,BVERSION,1,8)
        call TCPRECV(0,NP,4,1)
        call TCPRECV(0,NT,4,1)
        call TCPRECV(0,CSUM,4,1)
        call TCPRECV(0,SV,4,1)
      endif
C  Set values for other routines to use
      NPAR = NP
      NOBS = NT

C  Check for version skew
      if (AVERSION.ne.BVERSION) call BEOFATAL('BeoPEST versions differ.'
     |  //'  Expected '//TRIM(AVERSION)//' but got '//TRIM(BVERSION))

C  Allocate memory for buffers
      if (BEOMASTER) then
C       TCP manages memory internally - just need one buffer
        if (TCP) then
          N = 1
C       With MPI allocate memory for all nades
        else
          N = MAXNODE
        endif
C       Allocate memory
        ALLOCATE(PARBUF(NPAR,N),OBSBUF(NOBS,N),STAT=IOS)
        if (IOS.ne.0) call BEOFATAL('Cannot allocate buffer memory')
      endif
      END
C**********************************************************************C
C*******************  Push run from master to slave  ******************C
C**********************************************************************C
      SUBROUTINE BEOPUSH(IDST,PAR)

      use svd_parm_data, only : bnpar,BNSTMP,svalue,bitrans,sval,
     +                   last_nregadjpar,last_nlscaladj,
     +                   tbx,sf,svda_beoparset
      use pestdata, only : svda                                           !jd
      INCLUDE 'beopest.inc'
#ifdef BEOMPI
      INCLUDE 'mpif.h'
      INTEGER IOS
#endif
      INTEGER          IDST,N,J
      DOUBLE PRECISION PAR(NPAR),WALLTIME

C  Keep track of which slaves knows the new superparameters
      if (svda.eq.1 .and. svda_beoparset.eq.1) then
        super_needed = 1
        svda_beoparset = 0
      endif
C  Decide what to send
      if (svda.eq.1 .and. super_needed(IDST).eq.1) then
        N = 2
        super_needed(IDST) = 0
      else
        N = 1
      endif

C  Post asynchronous receive for parameters and targets
      if (TCP) then
        call TCPPOST(IDST,0,8,NPAR)
        call TCPPOST(IDST,1,8,NOBS)
      else
#ifdef BEOMPI
C  Post asynchronous receive for parameters and targets
        call MPI_IRECV(PARBUF(1,IDST),NPAR,MPI_DOUBLE_PRECISION,
     |     IDST,1,MPI_COMM_WORLD,IRDQ(1,IDST),IOS)
        call MPI_IRECV(OBSBUF(1,IDST),NOBS,MPI_DOUBLE_PRECISION,
     |     IDST,2,MPI_COMM_WORLD,IRDQ(2,IDST),IOS)
#endif
      endif

C  Mark start on this node
      if (TCP) LIVE(IDST) = 1
      STRT(IDST) = WALLTIME()
C  Send parameters to slave
      call BEOPUTINTS(IDST,N,1)
      call BEOPUTDOUBLES(IDST,PAR,NPAR)

C  Update SVD on slave (if necessary)
      if (N.eq.2) then
        call BEOPUTINTS(IDST,BNSTMP,1)
        call BEOPUTINTS(IDST,LAST_NREGADJPAR,1)
        call BEOPUTDOUBLES(IDST,SVALUE,NPAR)
        call BEOPUTDOUBLES(IDST,SVAL,BNPAR)
        call BEOPUTINTS(IDST,BITRANS,BNPAR)
        do j=1,BNSTMP
          call BEOPUTDOUBLES(IDST,tbx(1,j),NPAR-LAST_NREGADJPAR)
        enddo
        if (LAST_NLSCALADJ.ne.0) call BEOPUTDOUBLES(IDST,SF,BNPAR)
      endif

      return
      END
C**********************************************************************C
C******************  Wait for parameters from master  *****************C
C**********************************************************************C
      FUNCTION BEOGET(PAR)
      use pestdata, only : svda,apar,scale                                      !jd
      use svd_parm_data, only : bnpar,last_nregadjpar,BNSTMP,         !jd
     +                   last_nlscaladj,bntpfle,include_prior,svalue,           !jd
     +                   bapar,bitrans,sval,blbound,bubound,bscale,             !jd
     +                   boffset,SF,TBX                                                !jd

      INCLUDE 'beopest.inc'
#ifdef BEOMPI
      INCLUDE 'mpif.h'
      INTEGER STAT(MPI_STATUS_SIZE),IOS
#endif
      LOGICAL BEOGET
      INTEGER N,J,NREGADJPAR,NEEDED,IFAIL
      DOUBLE PRECISION PAR(NPAR)

C  Get count:  0 => done
      call BEOGETINTS(0,N,1)
C  Mark whether we got parameters or done
      BEOGET = (N.gt.0)
C  Return here if done
      if (.not.BEOGET) return

C  Get parameters from master
      call BEOGETDOUBLES(0,PAR,NPAR)

C  Get updated superparameters (if necessary)
      if (N.eq.2) then
C       Pause for paranoia
        if (svda.ne.1) call BEOFATAL('Slave not started with SVDA')
C       Read new Super-Base mapping from master
C       Do all reads right away to free up the master ASAP
C       Memory was allocated in INITIALISE_BASE
        call BEOGETINTS(0,BNSTMP,1)
        call BEOGETINTS(0,NREGADJPAR,1)
        call BEOGETDOUBLES(0,SVALUE,NPAR)
        call BEOGETDOUBLES(0,SVAL,BNPAR)
        call BEOGETINTS(0,BITRANS,BNPAR)
        do J=1,BNSTMP
          call BEOGETDOUBLES(0,TBX(1,J),NPAR-NREGADJPAR)
        enddo
        if (LAST_NLSCALADJ.ne.0) call BEOGETDOUBLES(0,SF,BNPAR)

C       Write a new parcalc.tpl
        write(6,'(/A)') ' BEOSLAVE writing new parcalc.tpl....'
        call write_parcalc_tpl(NREGADJPAR,LAST_NLSCALADJ,IFAIL)
        if (IFAIL.ne.0) call BEOFATAL('Failed to write parcalc.tpl')
        write(6,'(A/)') ' - file parcalc.tpl written ok.'
      endif

C  Return true (data read)
      return
      END
C**********************************************************************C
C**********************  Send targets to master  **********************C
C**********************************************************************C
      SUBROUTINE BEOPUT(PAR,OBS)
      INCLUDE          'beopest.inc'
#ifdef BEOMPI
      INCLUDE 'mpif.h'
      INTEGER IOS,I
#endif
      DOUBLE PRECISION PAR(NPAR),OBS(NOBS)

C  Send observations to master
      if (TCP) then
        call TCPSEND(0,PAR,8,NPAR)
        call TCPSEND(0,OBS,8,NOBS)
      else
#ifdef BEOMPI
        call MPI_SEND(PAR,NPAR,MPI_DOUBLE_PRECISION,
     |                0,1,MPI_COMM_WORLD,IOS)
        call MPI_SEND(OBS,NOBS,MPI_DOUBLE_PRECISION,
     |                0,2,MPI_COMM_WORLD,IOS)
#endif
      endif
      END
C**********************************************************************C
C**********************  Check for completed run  *********************C
C**********************************************************************C
      FUNCTION BEOSTAT(IDST)
      INCLUDE          'beopest.inc'
#ifdef BEOMPI
      INCLUDE 'mpif.h'
      LOGICAL FLAG
      INTEGER STAT(MPI_STATUS_SIZE),K,IOS
#endif
      LOGICAL          NEW
      INTEGER          BEOSTAT,IDST,IFLG
      DOUBLE PRECISION WALLTIME
C  RETURNS
C   1 done
C   0 working
C  -1 failed

      NEW = .false.
      if (TCP) then
C  Running
        if (LIVE(IDST).gt.0) then
C         Check if results were received
          call TCPTEST(IDST,IFLG)
C         Node has completed
          if (IFLG.gt.0) then
            NEW = .true.
            LIVE(IDST) = -1
C         Mark node as failed
          else if (IFLG.lt.0) then
            LIVE(IDST) = 0
          endif
C         Mark results as received
          BEOSTAT = IFLG
C  Idle => done
        else if (LIVE(IDST).lt.0) then
          BEOSTAT = 1
C  Dead
        else
          BEOSTAT = -1
        endif
      else
#ifdef BEOMPI
C  Check for outstanding receives that have completed
        do K=1,2
          if (IRDQ(K,IDST).ne.MPI_REQUEST_NULL) then
            call MPI_TEST(IRDQ(K,IDST),FLAG,STAT,IOS)
            if (FLAG) then
              NEW = .true.
              IRDQ(K,IDST) = MPI_REQUEST_NULL
            endif
          endif
        enddo
C  Return status
        if (IRDQ(1,IDST).eq.MPI_REQUEST_NULL .and.
     |      IRDQ(2,IDST).eq.MPI_REQUEST_NULL) then
          BEOSTAT = 1
        else
          BEOSTAT = 0
        endif
#else
        BEOSTAT = 0
#endif
      endif
      if (BEOSTAT.gt.0 .and. NEW) then
        NUM(IDST) = NUM(IDST)+1
        WALL(IDST) = WALLTIME() - STRT(IDST)
        TSUM = TSUM + WALL(IDST)
      endif
      END
C**********************************************************************C
C****************************  Get run time  **************************C
C**********************************************************************C
      FUNCTION BEOWALL(IDST)
      INCLUDE          'beopest.inc'
      INTEGER          IDST
      DOUBLE PRECISION BEOWALL

      BEOWALL = WALL(IDST)
      END
C**********************************************************************C
C**********************  Check for completed run  *********************C
C**********************************************************************C
      SUBROUTINE BEOFATAL(MSG)
      INCLUDE      'beopest.inc'
#ifdef BEOMPI
      INCLUDE 'mpif.h'
      INTEGER IOS
#endif
      CHARACTER(*) MSG

#ifdef BEOMPI
      if (MPI) call MPI_Finalize(IOS)
#endif
      call FATAL(TRIM(MSG)//CHAR(0))
      END
C**********************************************************************C
C**************  Return wall time in seconds since 1970  **************C
C**********************************************************************C
      FUNCTION WALLTIME()
      IMPLICIT NONE
      DOUBLE PRECISION WALLTIME
      INTEGER  d(8),MDO(12),JY,JD
      DATA MDO /306,337,0,31,61,92,122,153,184,214,245,275/

C  Get system date and time (wall clock local time)
      call DATE_AND_TIME(values=d)
C  Calculate sidereal Julian day for dates in Gregorian Calendar
      if (d(2).lt.3) then
        JY = d(1)-1
      else
        JY = d(1)
      endif
      JD = 1721119 + d(3) + MDO(d(2)) + 365*JY + JY/4 - JY/100 + JY/400
C  1-Jan-1970 is JD 2440588
C  Time is adjusted to UTC to avoid daylight savings glitches
      WALLTIME = 86400*(JD-2440588) + 3600*d(5) + (d(6)-d(4))*60
     |         + d(7) + 1D-3*d(8)
      END
C**********************************************************************C
C***********************  Return name of node  ************************C
C**********************************************************************C
      FUNCTION BEOTEXT(K)
      USE     BEOPEST
      INCLUDE 'beopest.inc'
      INTEGER       K
      CHARACTER*256 BEOTEXT

      BEOTEXT = TEXT(K)
      END
C**********************************************************************C
C***********************  Return name of node  ************************C
C**********************************************************************C
      SUBROUTINE BEOSTORE(NODE,JOB,PARREG,OBSREG,                           !jd
     +        irestart,ptunit)                                              !jd
      USE     PESTDATA,  ONLY: PARREGFILE,OBSREGFILE
      USE     BEOPEST
      INCLUDE 'beopest.inc'
      INTEGER NODE,JOB,PARREG,OBSREG,IOS,K,I
      integer irestart,ptunit                                               !jd
      integer ierr,jj                                                       !jd

C  Copy data from TCP buffer and store in register file
      if (TCP) then
        K = 1
        call TCPLOAD(NODE,0,PARBUF(1,K),8,NPAR)
        call TCPLOAD(NODE,1,OBSBUF(1,K),8,NOBS)
#ifdef BEOMPI
C  Set index to node
      else
        K = NODE
#endif
      endif
C  Store data to register files
      call STORE_PARALLEL_REGISTER(IOS,
     |  NPAR,PARREG,JOB,PARBUF(1,K),PARREGFILE)
      if (IOS.ne.0) call BEOFATAL('Cannot store returned PAR')
      call STORE_PARALLEL_REGISTER(IOS,
     |  NOBS,OBSREG,JOB,OBSBUF(1,K),OBSREGFILE)
      if (IOS.ne.0) call BEOFATAL('Cannot store returned OBJ')
      last_first_obs=obsbuf(1,k)                                            !jd

      if(irestart.ne.0)then                                                 !jd
        write(ptunit,iostat=ierr) job                                       !jd
        write(ptunit,iostat=ierr) (obsbuf(jj,k),jj=1,nobs)                  !jd
        if(ierr.ne.0) call BEOFATAL('Cannot save restart OBJ data.')        !jd
#ifdef FLUSHFILE
        call flush(ptunit)                                                  !jd
#endif
      end if                                                                !jd

      END


C**********************************************************************C
C***********************  Calculate average walltime ******************C
C**********************************************************************C

      double precision function average_time(new_time)

      use beopriv
      use pestdata, only : nespar
      implicit none

      double precision, intent(in) :: new_time

      integer i,n
      double precision dtemp

      if(nument_avbuf.eq.0)then
        nument_avbuf=max(nespar,10)
        ient_avbuf=0
        allocate(avbuf(nument_avbuf))
        avbuf=0      ! an array
      end if
      ient_avbuf=ient_avbuf+1
      if(ient_avbuf.gt.nument_avbuf)ient_avbuf=1
      avbuf(ient_avbuf)=new_time
      if(av_operating.eq.0)then
        if(ient_avbuf.gt.1)then
          av_operating=1
        end if
      end if

      dtemp=0.0d0
      n=0
      do i=1,nument_avbuf
        if(avbuf(i).ne.0.0d0)then
          n=n+1
          dtemp=dtemp+avbuf(i)
        end if
      end do
      if(dtemp.eq.0.0d0)then
        average_time=0.0d0
      else
        average_time=dtemp/n
      end if

      return
      end

C**********************************************************************C
C***********************  Send array of integers  *********************C
C**********************************************************************C
      SUBROUTINE BEOPUTINTS(IDST,K,N)
      INCLUDE 'beopest.inc'
#ifdef BEOMPI
      INCLUDE 'mpif.h'
      INTEGER IOS
#endif
      INTEGER IDST,K(N),N

      if (N.le.0) then
C       do nothing
      else if (TCP) then
        call TCPSEND(IDST,K,4,N)
      else
#ifdef BEOMPI
        call MPI_SEND(K,N,MPI_INTEGER,IDST,0,MPI_COMM_WORLD,IOS)
#endif
      endif
      END
C**********************************************************************C
C**********************  Receive array of integers  *******************C
C**********************************************************************C
      SUBROUTINE BEOGETINTS(ISRC,K,N)
      INCLUDE 'beopest.inc'
#ifdef BEOMPI
      INCLUDE 'mpif.h'
      INTEGER STAT(MPI_STATUS_SIZE),IOS
#endif
      INTEGER ISRC,K(N),N

      if (N.le.0) then
C       do nothing
      else if (TCP) then
        call TCPRECV(ISRC,K,4,N)
      else
#ifdef BEOMPI
        call MPI_RECV(K,N,MPI_INTEGER,ISRC,0,MPI_COMM_WORLD,STAT,IOS)
#endif
      endif
      END

C**********************************************************************C
C***********************  Send array of doubles  **********************C
C**********************************************************************C
      SUBROUTINE BEOPUTDOUBLES(IDST,X,N)
      INCLUDE 'beopest.inc'
#ifdef BEOMPI
      INCLUDE 'mpif.h'
      INTEGER IOS
#endif
      INTEGER          IDST,N
      DOUBLE PRECISION X(N)

      if (N.le.0) then
C       do nothing
      else if (TCP) then
        call TCPSEND(IDST,X,8,N)
      else
#ifdef BEOMPI
        call MPI_SEND(X,N,MPI_DOUBLE_PRECISION,IDST,0,
     |                MPI_COMM_WORLD,IOS)
#endif
      endif
      END
C**********************************************************************C
C**********************  Receive array of doubles  ********************C
C**********************************************************************C
      SUBROUTINE BEOGETDOUBLES(ISRC,X,N)
      INCLUDE 'beopest.inc'
#ifdef BEOMPI
      INCLUDE 'mpif.h'
      INTEGER STAT(MPI_STATUS_SIZE),IOS
#endif
      INTEGER          ISRC,N
      DOUBLE PRECISION X(N)

      if (N.le.0) then
C       do nothing
      else if (TCP) then
        call TCPRECV(ISRC,X,8,N)
      else
#ifdef BEOMPI
        call MPI_RECV(X,N,MPI_DOUBLE_PRECISION,ISRC,0,
     |                MPI_COMM_WORLD,STAT,IOS)
#endif
      endif
      END

